## TODO: Save obs percentiles

source("modules/Saving/R/get_dir.R")
source("modules/Saving/R/get_filename.R")
source("modules/Saving/R/Utils.R")
source("modules/Saving/R/save_forecast.R")
source("modules/Saving/R/save_observations.R")
source("modules/Saving/R/save_metrics.R")
source("modules/Saving/R/save_corr.R")
source("modules/Saving/R/save_probabilities.R")
source("modules/Saving/R/save_percentiles.R")

save_data <- function(recipe, data,
                      skill_metrics = NULL,
                      probabilities = NULL) {
  # Wrapper for the saving functions.
  # recipe: The auto-s2s recipe
  # archive: The auto-s2s archive
  # data: output of load_datasets()
  # data: output of calibrate_datasets()
  # skill_metrics: output of compute_skill_metrics()
  # probabilities: output of compute_probabilities()
  # mean_bias: output of compute_mean_bias()

  # Sanity checks
  if (is.null(recipe)) {
    error(recipe$Run$logger, "The 'recipe' parameter is mandatory.")
    stop()
  }

  if (is.null(data)) {
    error(recipe$Run$logger,
          paste("The 'data' parameter is mandatory. It should be a list",
               "of at least two s2dv_cubes containing the hcst and obs."))
    stop()
  }
  # Create output directory
  outdir <- get_dir(recipe)
  dir.create(outdir, showWarnings = FALSE, recursive = TRUE)

  # Export hindcast, forecast and observations onto outfile
  save_forecast(recipe = recipe, data_cube = data$hcst,
                type = 'hcst',
                outdir = outdir)
  if (!is.null(data$fcst)) {
    save_forecast(recipe = recipe, data_cube = data$fcst,
                  type = 'fcst',
                  outdir = outdir)
  }
  save_observations(recipe = recipe, data_cube = data$obs,
                   outdir = outdir)

  # Separate ensemble correlation from the rest of the metrics, as it has one
  # extra dimension "ensemble" and must be saved to a different file
  if ("corr" %in% names(skill_metrics)) {
    corr_metric_names <- grep("^corr", names(skill_metrics))
    corr_metrics <- skill_metrics[corr_metric_names]
    skill_metrics <- skill_metrics[-corr_metric_names]
    if (length(skill_metrics) == 0) {
      skill_metrics <- NULL
    }
  } else {
    corr_metrics <- NULL
  }

  # Export skill metrics onto outfile
  if (!is.null(skill_metrics)) {
    save_metrics(recipe = recipe, skill = skill_metrics,
                 data_cube = data$hcst,
                 outdir = outdir)
  }
  if (!is.null(corr_metrics)) {
    save_corr(recipe = recipe, skill = corr_metrics,
              data_cube = data$hcst,
              outdir = outdir)
  }
  
  # Export probabilities onto outfile
  if (!is.null(probabilities)) {
    save_percentiles(recipe = recipe, percentiles = probabilities$percentiles,
                     data_cube = data$hcst,
                     outdir = outdir)
    save_probabilities(recipe = recipe, probs = probabilities$probs,
                       data_cube = data$hcst,
                       type = "hcst",
                       outdir = outdir)
    if (!is.null(probabilities$probs_fcst)) {
      save_probabilities(recipe = recipe, probs = probabilities$probs_fcst, 
                         data_cube = data$fcst,
                         type = "fcst",
                         outdir = outdir)
    }
  }
}

get_global_attributes <- function(recipe, archive) {
  # Generates metadata of interest to add to the global attributes of the
  # netCDF files.
  parameters <- recipe$Analysis
  hcst_period <- paste0(parameters$Time$hcst_start, " to ",
                        parameters$Time$hcst_end)
  current_time <- paste0(as.character(Sys.time()), " ", Sys.timezone())
  system_name <- parameters$Datasets$System$name
  reference_name <- parameters$Datasets$Reference$name

  attrs <- list(reference_period = hcst_period,
                institution_system = archive$System[[system_name]]$institution,
                institution_reference = archive$Reference[[reference_name]]$institution,
                system = system_name,
                reference = reference_name,
                calibration_method = parameters$Workflow$Calibration$method,
                computed_on = current_time)
  
  return(attrs)
}

get_times <- function(store.freq, fcst.horizon, leadtimes, sdate, calendar) {
  # Generates time dimensions and the corresponding metadata.
  ## TODO: Subseasonal

  switch(fcst.horizon,
         "seasonal" = {time <- leadtimes; ref <- 'hours since '; 
                       stdname <- paste(strtoi(leadtimes), collapse=", ")},
         "subseasonal" = {len <- 4; ref <- 'hours since ';
                          stdname <- ''},
         "decadal" = {time <- leadtimes; ref <- 'hours since ';
                      stdname <- paste(strtoi(leadtimes), collapse=", ")})

  dim(time) <- length(time)
  sdate <- as.Date(sdate, format = '%Y%m%d') # reformatting
  metadata <- list(time = list(units = paste0(ref, sdate, 'T00:00:00'),
                   calendar = calendar))
  attr(time, 'variables') <- metadata
  names(dim(time)) <- 'time'
  
  sdate <- 1:length(sdate)
  dim(sdate) <- length(sdate)
  metadata <- list(sdate = list(standard_name = paste(strtoi(sdate),
                                                      collapse=", "),
                                units = paste0('Init date')))
  attr(sdate, 'variables') <- metadata
  names(dim(sdate)) <- 'sdate'

  return(list(time=time))
}

get_latlon <- function(latitude, longitude) {
  # Adds dimensions and metadata to lat and lon
  # latitude: array containing the latitude values
  # longitude: array containing the longitude values
  
  dim(longitude) <- length(longitude)
  metadata <- list(longitude = list(units = 'degrees_east'))
  attr(longitude, 'variables') <- metadata
  names(dim(longitude)) <- 'longitude'
  
  dim(latitude) <- length(latitude)
  metadata <- list(latitude = list(units = 'degrees_north'))
  attr(latitude, 'variables') <- metadata
  names(dim(latitude)) <- 'latitude'

  return(list(lat=latitude, lon=longitude))

}
